<?php
namespace APS;

require_once "aps/2/json.php";
require_once "aps/2/utils.php";

class AnnotationUtils {

	public static function generateSimpleAnnotation($object, $annotationName) {
		if(isset($object->$annotationName))
			return " * @$annotationName\n";
		return '';
	}

	public static function generateValuedAnnotation($object, $annotationName) {
		if(isset($object->$annotationName))
			return " * @$annotationName(". $object->$annotationName .")\n";
		return '';
	}  
	
	public static function generateQuotedAnnotation($object, $annotationName) {
		if(isset($object->$annotationName))
			return " * @$annotationName(\"". $object->$annotationName ."\")\n";
		return '';		
	}

	public static function generateSmartAnnotation($object, $annotationName) {
		if(isset($object->$annotationName))
			return $object->$annotationName->toSource();
		return '';		
	}
}

/** 
 * get annotation from one line of comment
 * @param string $line
 * @return array($name, $p1, $p2, ...)
 */
function getAnnotation($line)
{
	$line = preg_replace('/^\s*\*\s*/', '', $line);
	
	// @annotation(p) 
	// @annotation("p")
	// @annotation('p')
	// @annotation(p[])
	// @annotation(p1, "p2", 'p3', p4[] ...)
	if (preg_match('/@(\w+)\((.*\S.*)\)[\s;]*$/', $line, $matches)) {
		$result = array($matches[1]);
		$annotations = $matches[2];
		// retrieve parts separated with comma
		while (preg_match('/^(\s*(([\w\/\\\\]+(\[\])?)|"([^"]*?)"|\'([^\']*?)\')\s*(,|$))/', $annotations, $matches)) {
			$matched = $matches[1];
			array_push($result, preg_replace('/["\']$/', '', preg_replace('/^["\']/', '', $matches[2])));
			$annotations = substr($annotations, strlen($matched));
		}
		if (count($result) == 1) {
			throw new \Exception("Invalid format of annotation $line");
		}
		return $result;
	}

	// @annotation
	// @annotation()
	if (preg_match('/@(\w+)(\(\s*\))?\s*$/', $line, $matches))
	{
		return array($matches[1]);
	}
	return NULL;
}

/** 
 * Parse comment for annotatations
 * @param ReflectionObject $refl
 * @return array($name, $p1, $p2, ...)[]
 */
function getAnnotations($refl)
{
	$comment = $refl->getDocComment();
	$comment = substr($comment, 3, -2); // cut comment signs
	$comment = trim($comment); 
	
	$ret = array();	
	foreach(explode("\n", $comment) as $line)
	{	
		$ann = getAnnotation($line);
		if($ann)
			$ret[] = $ann;
	}
	return $ret;
}

class SchemaException extends \Exception 
{
}

class Access
{
	public $owner;
	public $admin;
	public $referrer;
	
	protected static $serializable = array('owner', 'admin', 'referrer');

	public function __construct($obj = null)
	{
		if ($obj)
		{
			// init by serialized object
			foreach(self::$serializable as $k)
				if (property_exists($obj, $k))
					$this->$k = $obj->$k;
		}
	}

	static public function fillAccessByAnnotation($obj, $name, $val)
	{
		if (!$obj->access)
			$obj->access = new Access();

		switch($name) 
		{
			case "owner":
			case "admin":
			case "referrer":
				if ($val == "true")
					$obj->access->$name = true;
				elseif ($val == "false")
					$obj->access->$name = false;
				else
					$obj->access->$name = $val;
		}	
	}
	
	public function toSource() 
	{
		$ret = "";
		foreach (self::$serializable as $k)
			if (isset($this->$k))
				$ret .= " * @access(".$k.",".($this->$k ? "true" : "false").")\n";
		return $ret;
	}
	
	public function toArray()
	{
		$ret = new \stdClass();
		$n = 0;
		foreach (self::$serializable as $k)
			if (isset($this->$k))
			{
				$ret->$k = (bool)$this->$k;
				$n++;
			}
			
		return $n ? $ret : null;
	}
}

/**
 * Holder of typed value
 */
class TypedValue
{
	public $type;
	public $items;
	
	/** 
	 * True if attribute holds array of types 
	 * @return bool 
	 */
	public function isArray() {
		return !is_null($this->items);
	}
	
	/** 
	 * Get type of value (simple or complex), array never returned here
	 * @return string
	 */
	public function getType()
	{
		if(isset($this->items))
			return $this->items->type;
		return $this->type;
	}
	
	/** Is attribute of simple type */
	public function isSimple()
	{
		return array_key_exists($this->getType(), Schema::$simpleTypes);
	}
	
	/** Parse type value form form 'typeName' or 'typeName[]' */
	public function parseType($typeText)
	{
		if(substr($typeText, -2, 2) == '[]')
		{
			$this->type = 'array';
			$this->items = new \stdClass();
			$this->items->type = substr($typeText, 0, -2);
		}
		else
		{
			$this->type = $typeText;
			$this->items = NULL;
		}
	}
	
	public function __construct($typeText = NULL)
	{
		if ($typeText)
			$this->parseType($typeText);
	}
	/**
	 * Generate source for type annotation
	 * Example:
	 * @type(string) for single values
	 * @type(string[]) for arrays
	 */
	public function toSource() {
		return " * @type(".$this->typeToSource().")\n";
	}
	
	public function addNSToType($namespace) {
		if (!$namespace)
			return;
			
		if ($this->isSimple())
			return;
			
		if (strpos($this->getType(), "://"))
			return;
			
		if(is_null($this->items))
			$this->type = $namespace.'\\'.$this->type;
		else
			$this->items->type = $namespace.'\\'.$this->items->type;
	}
	
	public function typeToSource()
	{
		$type = is_null($this->items) ? $this->type : $this->items->type.'[]';
		$type = (strpos($type, '://') === False) ? $type : "\"$type\"";
		return $type;
	}
}

class TypedMediaValue extends TypedValue
{
	public $contentType;
}

/** 
 * APS Interface Type property class
 */
class Property extends TypedValue
{
	public $name;
	public $required;
	public $readonly;
	public $final;
	public $encrypted;
	public $default;
	public $headline;
	public $format;
	public $pattern;
	public $title;
	public $description;
	public $unit;
	public $minLength;
	public $maxLength;
	public $minItems;
	public $maxItems;
	public $uniqueItems;
	public $enum;
	public $enumTitles;
	public $access;

	protected static $serializable = array(
			'type',
			'items',
			'required', 
			'readonly',
			'final',
			'encrypted',
			'default',
			'headline',
			'format',
			'pattern',
			'title',
			'description',
			'unit',
			'minLength',
			'maxLength',
			'minItems',
			'maxItems',
			'uniqueItems',
			'enum',
			'enumTitles');
	
	/**
	 * Create Attribute either from DTO structure or from ReflectionProperty
	 * @param mixed $obj - ReflectionProperty or object from json_decode()
	 * @param string $name - property name (not assigned if missed)
	 */ 
	public function __construct($obj, $namespace = '', $name = NULL)
	{
		if ($obj instanceof \ReflectionProperty)
		{
			// init by reflection object
			$this->scan($obj);
		}
		else
		{
			// init by serialized object
			foreach(self::$serializable as $k)
				if (property_exists($obj, $k))
					$this->$k = $obj->$k;

			if ($name)
				$this->name = $name;
				
			if (isset($obj->access))
				$this->access = new Access($obj->access);
		}

		$this->addNSToType($namespace);
	}
	
	/**
	 * Scan one line of annotation
	 * @param array $ann - parsed annotation array
	 */
	public function scanAnnotation($ann)
	{
		$name = array_shift($ann);
		switch($name) 
		{
			case "type":
				$this->parseType($ann[0]);
				break;
			case "required":
			case "readonly":
			case "final":
			case "encrypted":
			case "uniqueItems":
			case "headline":
				$this->$name = true;
				break;
			case "format":
			case "pattern":
			case "title":
			case "description":
				$this->$name = $ann[0];
				break;
			case "unit":
				$this->$name = $ann[0];
				break;
			case "minLength":
			case "maxLength":			
			case 'minItems':
			case 'maxItems':
				$this->$name = (int)$ann[0];
				break;
			case "option":
				$this->enum[] = $ann[0];
				if(isset($ann[1]))
					$this->enumTitles[] = $ann[1];
				break;
			case "access":
					Access::fillAccessByAnnotation($this, $ann[0], $ann[1]);
					break;
		}
		return true;	
	}
	
	/**
	 * Scan parameter for annotations
	 * @param ReflectionProperty $prop
	 * @throws SchemaException
	 */
	public function scan($prop)
	{
		$this->name = $prop->getName();
		foreach(getAnnotations($prop) as $ann)
			$this->scanAnnotation($ann);

		$defaults = $prop->getDeclaringClass()->getDefaultProperties();
		if (key_exists($this->name, $defaults))
			$this->default = $defaults[$this->name];
			
		if (!$this->type)
			throw new SchemaException("Can't parse ".$this->name." - type is not defined"); 
		
 	}
 	
 	/**
 	 * Serialize method to sutructure suitable for json_encode()
 	 */
 	public function toArray()
 	{
 		$attrs = array();
		foreach (self::$serializable as $k)
			if (!is_null($this->$k)) {
				$value = $this->$k;
				
				if($k == 'type' && ($shift = strrpos($this->$k, '\\')))
					$value = substr($this->$k, $shift + 1);
					
				if($k == 'items' && ($shift = strrpos($this->$k->type, '\\'))) {
					$value = new \stdClass();
					$value->type = substr($this->$k->type, $shift + 1);
				}
				
				$attrs[$k] = $value;
			}
			
		if ($this->access)
			$attrs["access"] = $this->access->toArray();
				
		return $attrs;
 	}
 	
	public function createEnumAnnotation() {
		$enumAnnotation = '';
		if(isset($this->enum))
			foreach($this->enum as $key => $enum) {
				$enumTitle = isset($this->enumTitles[$key]) ? ', "'.$this->enumTitles[$key].'"' : '';
				$enumAnnotation .= " * @option(\"$enum\"$enumTitle)\n";
			}
		return $enumAnnotation;
	}
	
	public function escapeSourceLiteral($value)
	{
		switch($this->getType())
		{
			case Schema::TYPE_STRING:
				return '"' . $value . '"';
			// Need process escaping?
				
			case Schema::TYPE_INTEGER:
			case Schema::TYPE_NUMBER:
				return "".$value;
			
			case Schema::TYPE_BOOLEAN:
				return $value ? 'true' : 'false';

		}	
	}
	
 	//TODO: consider unquoted values 
	public function getDefaultValue() {
		if (!$this->isSimple())
			return '';
			
		$t = $this->getType();
		if(is_array($this->default)) 
		{
			$defaults = array();
			foreach($this->default as $value) 
				$defaults[] = $this->escapeSourceLiteral($value);

			return " = array(".implode(", ", $defaults).")";
		}
		elseif(isset($this->default)) 
		{
			return " = ". $this->escapeSourceLiteral($this->default);
		}
		return '';	
	}

 	public function toSource() {

 		return "/**\n" .
 				parent::toSource() .
 		AnnotationUtils::generateSimpleAnnotation($this, 'required') .
 		AnnotationUtils::generateSimpleAnnotation($this, 'readonly') .
 		AnnotationUtils::generateSimpleAnnotation($this, 'final') .
 		AnnotationUtils::generateSimpleAnnotation($this, 'encrypted') .
 		AnnotationUtils::generateSimpleAnnotation($this, 'uniqueItems') .
 		AnnotationUtils::generateSimpleAnnotation($this, 'headline') .
 		AnnotationUtils::generateQuotedAnnotation($this, 'format') .
 		AnnotationUtils::generateQuotedAnnotation($this, 'pattern') .
 		AnnotationUtils::generateQuotedAnnotation($this, 'title') .
 		AnnotationUtils::generateQuotedAnnotation($this, 'description') .
 		AnnotationUtils::generateQuotedAnnotation($this, 'unit') .
 		AnnotationUtils::generateValuedAnnotation($this, 'minLength') .
 		AnnotationUtils::generateValuedAnnotation($this, 'maxLength') .
 		AnnotationUtils::generateValuedAnnotation($this, 'minItems') .
 		AnnotationUtils::generateValuedAnnotation($this, 'maxItems') .
 		AnnotationUtils::generateSmartAnnotation ($this, 'access') .
 		$this->createEnumAnnotation() .
 				" */\n".
 				"public $$this->name".($this->getDefaultValue()).";\n";
 	}
}


/** 
 * APS Interface Attribute class
 */
class Attribute extends Property
{
	private $link;

	/** Is attribute is link to other object */
	public function isLink()
	{
		return (bool)$this->link;
	}
		
	/** Create Attribute either from DTO structure or from ReflectionProperty */
	public function __construct($obj, $namespace = '', $name = NULL)
	{
		if ($obj instanceof \ReflectionProperty)
		{
			// init by reflection object
			$this->scan($obj);
		}
		else
		{
			// init by serialized object
			foreach(self::$serializable as $k)
				if (property_exists($obj, $k))
					$this->$k = $obj->$k;
					
				if ($name)
					$this->name = $name;
					
				if (isset($obj->access))
					$this->access = new Access($obj->access);
		}

		$this->addNSToType($namespace);
	}
	
	public function scan($prop)
	{
		$this->name = $prop->getName();
		foreach(getAnnotations($prop) as $ann)
		{
			parent::scanAnnotation($ann);
			switch($ann[0]) 
			{
				case "link":
						$type = $ann[1]; // $type is unused here
						$this->link = true;
						break;
			}
		}
		if (!$this->type)
			throw new SchemaException("Can't parse $this->name - type is not defined"); 
		
		$defaults = $prop->getDeclaringClass()->getDefaultProperties();
		if (key_exists($this->name, $defaults))
			$this->default = $defaults[$this->name];
 	}
 	
 	public function toArray()
 	{
 		$attrs = array();
		foreach (self::$serializable as $k)
			if (!is_null($this->$k)) {
				$value = $this->$k;
				
				if($k == 'type' && ($shift = strrpos($this->$k, '\\')))
					$value = substr($this->$k, $shift + 1);
					
				if($k == 'items' && ($shift = strrpos($this->$k->type, '\\'))) {
					$value = new \stdClass();
					$value->type = substr($this->$k->type, $shift + 1);
				}
				
				$attrs[$k] = $value;
			}
			
		if ($this->access)
			$attrs["access"] = $this->access->toArray();

		return $attrs;
 	}
}

/** Method parameters holder class */
class Parameter extends TypedValue
{
	public $name;
	public $index;
	public $required;
	public $default;
	public $kind;
	
	public function toSource() {
		return " * @param(".$this->typeToSource().", $this->kind)\n";
	}
	
	public static $serializable = array('type', 'items', 'required', 'kind');	
}

/** APS Interface Operation class */
class Operation
{
	public $name;
	public $description;
	public $verb;
	public $path;
	public $response;
	public $parameters;
	public $static;
	public $access;
	
	private static $serializable = array('description', 'verb', 'path', 'response', 'static');
	
	/** Is Array returned */
	public function isArrayReturned()
	{
		return(substr($this->returns, -2, 2) == '[]');
	}
	
	/**
	 * Build operation on base of either 
	 * @param mixed $method - either ReflectionMethod or operation method from json_decode()
	 */
	public function __construct($method, $name = NULL)
	{
		if ($method instanceof \ReflectionMethod)
		{
			$this->scan($method);
		}
		else 
		{
			// init by serialized object
			foreach(self::$serializable as $k)
				if (property_exists($method, $k))
					$this->$k = $method->$k;
					
			if (property_exists($method, 'parameters'))
			{
				foreach($method->parameters as $k => $p)
				{
					$pnew = new Parameter();
					$pnew->name = $k;
					foreach(Parameter::$serializable as $pk)
						if (property_exists($p, $pk))
							$pnew->$pk = $p->$pk;
					$this->parameters[] = $pnew;
				}
			}
			if($name)
				$this->name = $name;
				
			if (isset($method->access))
				$this->access = new Access($method->access);
		}
	}
	
	/**
	 * Scan method for annotations
	 * @param ReflectionMethod $method
	 * @throws SchemaException
	 */
	public function scan($method)
	{
		$this->name = $method->getName();

		if ($method->isConstructor() || $method->isDestructor() || 
			$method->isPrivate() || $method->isProtected () ||
				$method->isStatic())
					throw new SchemaException("Method $this->name is nut suitable for binding");

		$this->path = "/";
		$pmeta = array();
		foreach(getAnnotations($method) as $ann)
		{
			switch($ann[0]) 
			{
				case "verb": 
						if (!in_array($ann[1], array( 'GET', 'POST', 'PUT', 'DELETE' )))
							throw new SchemaException("Unknown http verb '$ann[1]'");
	
						$this->verb = $ann[1];
						break;
						
				case "path": 
						$this->path = $ann[1];
						break;
						
				case "return":
						$this->response = new TypedMediaValue($ann[1]);
						if(isset($ann[2]))
							$this->response->contentType = $ann[2];
						break;
						
				case "description":
						$this->description = $ann[1];
						break;

				case "param":
						$pmeta[] = array_splice($ann, 1);
						break;

				case "static":
						$this->static = true;
						break;
						
				case "access":
						Access::fillAccessByAnnotation($this, $ann[1], $ann[2]);
						break;
			}
		}
		if (!$this->verb)
			throw new SchemaException("Method $this->name not exposed");
			
		$this->parameters = array();
		
		foreach ($method->getParameters() as $p)
		{
			$t = new TypedValue();
			if (count($pmeta) <= $p->getPosition()) {
				throw new \Exception("Can not identify type of parameter '".$p->name."' of method '".$method->name."'! Method has more arguments than @param annotaions!");
			}
			$t->parseType($pmeta[$p->getPosition()][0]);
			$prm = new Parameter();
			$prm->name = $p->getName();
			$prm->index = $p->getPosition();
			$prm->type = $t->type;
			if ($t->isArray())
				$prm->items = $t->items;
			if (isset($pmeta[$prm->index][1]))
				$prm->kind = $pmeta[$prm->index][1];
			elseif ($prm->type == 'self' or $prm->type == 'this')
				$prm->kind = 'body';
			else
				$prm->kind = 'query';
			$prm->required = !$p->isOptional();
			if ($p->isOptional())
				$prm->default = $p->getDefaultValue();
			$this->parameters[] = $prm;
		}		
 	}
 	
 	/**
 	 * Serialize method to sutructure suitable for json_encode()
 	 */
 	public function toArray()
 	{
 		$ops = array();
		foreach (self::$serializable as $k)
			if ($this->$k)
				$ops[$k] = $this->$k;
			
		if ($this->access)
			$ops["access"] = $this->access->toArray();
			
			
 		if ($this->parameters)
 		{
			$ops['parameters'] = array();
			foreach ($this->parameters as $p)
			{
				$prm = array();
				foreach (Parameter::$serializable as $k)
				if (isset($p->$k))
						$prm[$k] = $p->$k;
				$ops['parameters'][$p->name] = $prm;
			}
 		}
		return $ops;
 	}
 	
 	public function responseToSource() {
 		$response = $this->response;
 		if(!is_null($response)) {
	 		$contentType = isset($response->contentType) ? ", $response->contentType" : '';
	 		$isArray = isset($response->items) ? '[]': '';
	 		$type = isset($response->items) ? $response->items->type : $response->type;
	 		return " * @return(". $type . $isArray .$contentType .")\n";
 		}
 		return '';
 	}
 	
 	public function toSource() {
 		$operationAnnotation = "/**\n" .
 		AnnotationUtils::generateQuotedAnnotation($this, 'description') . 
 		AnnotationUtils::generateValuedAnnotation($this, 'verb') . 
 		AnnotationUtils::generateQuotedAnnotation($this, 'path') . 
 		AnnotationUtils::generateSimpleAnnotation($this, 'static') .
 		AnnotationUtils::generateSmartAnnotation ($this, 'access') .
 		$this->responseToSource();
 		if(!is_null($this->parameters))
			foreach($this->parameters as $parameter) {
				$operationAnnotation .= $parameter->toSource();
			}
		$operationAnnotation .= " */\n";
 		$arguments = array();
 		$bodyPresent = false;
 		if(!is_null($this->parameters))
	 		foreach($this->parameters as $parameter) {
	 			$arguments[] = "$$parameter->name" . (isset($parameter->default) ? " = $parameter->default" : '');
	 			if($parameter->kind === 'body') {
	 				$bodyPresent = true;
	 				$input = "$$parameter->name";
	 			}
	 		}
 		$argumentString = implode(', ', $arguments);
		return "\n${operationAnnotation}public function $this->name($argumentString) {\n\treturn \\APS\\Request::getController()->getIo()->sendRequest(\"$this->verb\", \\APS\\Proto::resourcePath(\$this->aps->id, \"$this->path\")" .
 		($bodyPresent ? ", $input" : '') . ");\n}\n";
 	}

	public function applyArguments($args) {
		$methodPath = $this->path;
		$methodBody = null;
		if(!is_null($this->parameters))
			for ($i=0; $i < count($this->parameters); ++$i) {
				$parameter = $this->parameters[$i];
				$param_value = isset($parameter->default) ? $parameter->default : "";
				if (array_key_exists($i, $args)) {
					$param_value = $args[$i];
				}
				if($parameter->kind === 'body') 
					$methodBody = $param_value;
				else
					$methodPath = str_replace('{'.$parameter->name.'}', $param_value, $methodPath);
			}
		return array($methodPath, $methodBody);
	}
}

/** 
 * APS Interface Type class
 */
class Type
{
	public $name;
	public $description;
	public $properties;
	
	private static $serializable = array('name', 'description');
	
	public function getProperty($name)
	{
		if (!array_key_exists($name, $this->properties))
			throw new SchemaException("Unknown property '$name' in type $this->name");
		return $this->properties[$name];
	}
	
	/** return list of properties */
	public function getProperties()
	{
		return $this->properties;
	}
	
	/**
	 * Create Type either from DTO structure or from ReflectionClass
	 * @param mixed $obj - ReflectionClass or object from json_decode()
	 */ 
	public function __construct($obj, $namespace = '', $name = NULL)
	{
		if ($obj instanceof \ReflectionClass)
		{
			// init by reflection object
			$this->scan($obj, $namespace);
		}
		else
		{
			// init by serialized object
			foreach(self::$serializable as $k)
				if (property_exists($obj, $k))
					$this->$k = $obj->$k;
					
			if(property_exists($obj, 'properties'))
				foreach($obj->properties as $k => $p)
					$this->properties[$k] = new Property($p, $namespace, $k);
			
			if ($name)
				$this->name = $name;
		}
	}
	
	/**
	 * Scan Type for annotations
	 * @param ReflectionClass $refl
	 */
	public function scan($refl, $namespace)
	{
		$this->name = $refl->getName();
		foreach(getAnnotations($refl) as $ann)
		{
			switch($ann[0]) 
			{
				case "description":
						$this->description = $ann[1];
						break;
			}
		}
		
		$this->properties = array();
		foreach($refl->getProperties() as $prop)
		{
			try
			{
				$k = $prop->getName();
				$p = new Property($prop, $namespace);
				$this->properties[$k] = $p;
			}
			catch(SchemaException $ex)
			{
			}
		}
 	}
 	
 	/**
 	 * Serialize method to sutructure suitable for json_encode()
 	 */
 	public function toArray()
 	{
 		$attrs = array();
		foreach (self::$serializable as $k)
			if (isset($this->$k))
				$attrs[$k] = $this->$k;

		$attrs['type'] = 'object';
		if ($this->properties) {
			$props = array();			
			foreach($this->properties as $k =>$p)
				$props[$k] = $p->toArray();
			$attrs['properties'] = $props;
		}
		return $attrs;
 	}
 	
 	public function toSource() {
 		$properties = '';
 		if($this->properties)
	 		foreach($this->properties as $pname => $property)
	 			$properties .= $property->toSource();
 		$shortName = $this->name;
 		if($offset = strrpos($this->name, '\\'))
 			$shortName = substr($this->name, $offset + 1);
 		return "class $shortName {\n" . $properties . "\n}\n";
 	}
}


/** 
 * APS Interface Type relation class
 */
class Relation extends TypedValue
{
	public $name;
	public $required;
	public $requirement;
	public $collection;
	public $allocate;
	public $assign;
	
	private static $serializable = array(
			'type',
			'required', 
			'requirement', 
			'collection',
			'assign',
			'allocate'
	);
	
	public function isArray()
	{
		return $this->collection;
	}
	
	/**
	 * Create Relation either from DTO structure or from ReflectionProperty
	 * @param mixed $obj - ReflectionProperty or object from json_decode()
	 * @param string $name - property name (not assigned if missed)
	 */ 
	public function __construct($obj, $name = NULL)
	{
		if ($obj instanceof \ReflectionProperty)
		{
			// init by reflection object
			$this->scan($obj);
		}
		else
		{
			// init by serialized object
			foreach(self::$serializable as $k)
				if (property_exists($obj, $k))
						$this->$k = $obj->$k;
			
			if ($name)
				$this->name = $name;
		}
	}
	
	/**
	 * Scan one line of annotation
	 * @param array $ann - parsed annotation array
	 */
	public function scanAnnotaion($ann)
	{
		$name = array_shift($ann);
		switch($name) 
		{
			case "link":
					$this->parseType($ann[0]);
					$this->collection = ($this->type == 'array');
					if ($this->collection)
					{
						$this->type = $this->items->type;
						unset($this->items);
					}
					break;
		
			case "requirement":
					if ($this->requirement)
						foreach ($ann as $annotation)
							$this->requirement .= ", " . $annotation;
					else 
						$this->requirement = $ann[0];
					break;
					
			case "required":
					$this->required = true;
					break;
				
			case "allocate":
				$this->allocate = $ann[0];
				break;
			case "assign":
					$this->assign[] = $ann[0];
					break;
					
			default:
				return false;
		}
		return true;	
	}
	
	/**
	 * Scan parameter for annotations
	 * @param ReflectionProperty $prop
	 * @throws SchemaException
	 */
	public function scan($prop)
	{
		$this->required = null;
		$this->name = $prop->getName();
		foreach(getAnnotations($prop) as $ann)
			$this->scanAnnotaion($ann);

		if (!$this->type)
			throw new SchemaException("Can't parse ".$this->name." - type is not defined"); 
		
 	}
 	
 	/**
 	 * Serialize method to sutructure suitable for json_encode()
 	 */
 	public function toArray()
 	{
 		$attrs = array();
		foreach (self::$serializable as $k)
			if (isset($this->$k))
				$attrs[$k] = $this->$k;
		return $attrs;
 	}
 	
 	public function toSource()
 	{
 		$multiLineRequirement = '';
 		if(isset($this->requirement)) {
	 		$requirements = explode(", ", $this->requirement);
	 		foreach($requirements as $requirement) {
	 			$multiLineRequirement .= " * @requirement(\"$requirement\")\n";
	 		}
 		}
 			
 		$multiLineAssign = '';
 		if(isset($this->assign))
	 		foreach($this->assign as $assign) {
	 			$multiLineAssign .= " * @assign(\"$assign\")\n";
	 		}
 	
		$collection_sign = '';
		if(isset($this->collection) && $this->collection) {
			$collection_sign = '[]';
		}
 			
 		return "/**\n" .
 				AnnotationUtils::generateSimpleAnnotation($this, 'required') .
 				AnnotationUtils::generateQuotedAnnotation($this, 'allocate') .
 				$multiLineRequirement .
 				" * @link(\"$this->type$collection_sign\")\n" .
 				$multiLineAssign .
 				" */\n".
 				"public $$this->name;\n";
 	}
}


/** APS Interface descriptor class */
class Schema
{
	const TYPE_STRING 	= "string";
	const TYPE_INTEGER 	= "integer";
	const TYPE_NUMBER 	= "number";
	const TYPE_BOOLEAN 	= "boolean";

	const TOP_RESOURCE	= "http://aps-standard.org/types/core/application";
	
	static public $simpleTypes = array(
		self::TYPE_STRING   => "Text string, utf-8 characters are allowed",
		self::TYPE_INTEGER  => "An integer number, 64bit long",
		self::TYPE_NUMBER   => "A float number, double precision", 
		self::TYPE_BOOLEAN  => "Boolean type (true or false)"
	);
	
	
	var $apsVersion = "2.0";

	/** 
	 * Class name (short), fits to php 
	 * @var string 
	 */
	var $name;
	
	/** 
	 * Interface ID, unique defines type for class in APS namespace
	 * @var string 
	 */
	var $id;
		
	/** 
	 * List of implemented interfaces
	 * @var string[]
	 */
	var $implements = array();
	
	/** 
	 * List of properties
	 * @var Attribute[]
	 */
	var $properties = array();
	
	/** 
	 * List of operations
	 * @var Operation[]
	 */
	var $operations = array();

	/** 
	 * List of structures used by type
	 * @var Type[]
	 */
	var $structures = array();

	/** 
	 * List of implicetely defined structures (may be not used)
	 * @var Type[]
	 */
	var $defStructures = array();
	
	/**
	 * List of relations used by type
	 * @var Relations[]
	 */
	var $relations = array();
	
	/**
	 * Service name for the schema
	 * @var service
	 */
	var $service;
	
	/**
	 * Base schema (schema defined in one file without resolved inheretance)
	 * @var $base
	 */
	public $base;
	
	/**
	 * Access structure
	 */
	public $access;

	/**
	 * Top level service - the one which implements core/application
	 * @var $isTopResource
	 */
	public $isTopResource = false;

	/** return attribute(property) by name */
	public function getProperty($name)
	{
		if (!array_key_exists($name, $this->properties))
			throw new SchemaException("Unknown attribute '$name' in type $this->type");
		return $this->properties[$name];
	}
	
	/** return list of properties(properties) */
	public function getProperties()
	{
		return $this->properties;
	}
	

	/** return list of auxilarily structures */
	public function getStructures()
	{
		return array_merge($this->structures, $this->defStructures);
	}

	/**
	 * Construct schema class
	 * @param array $prms - key => value constructor parameters
	 * Construct by class scanning:
	 * 		$schema = new Schema(array('class' => 'MyClassNameHere'));
	 * 		class McClassNameHere should be defined in PHP scome and annotated properly
	 * Construct by serialized schema (JSON format):
	 * 		$schema = new Schema(array('schema' => '{"apsVersion": "2.0", ....'));
	 * 		Serialize schema should match APS Type Schema.
 	 */
	public function __construct($prms = NULL)
	{
		if($prms && isset($prms['schema']))
			$this->fromString($prms['schema']);
		elseif ($prms && isset($prms['class']))
			$this->scan($prms['class']);
		elseif($prms && isset($prms['name']) && isset($prms['src']))
			$this->fromSource($prms['src'], $prms['name']);
	}

	/**
	 * Scan for properties
	 * @param ReflectionClass $refl - object to scan for propertyies annotations
	 */
	 private function scanProperties($refl)
	 {
	 	$this->properties = array();
	 	foreach ($refl->getProperties() as $prop)
		{
			try
			{
				$k = $prop->getName();
				$a = new Attribute($prop);
				$this->properties[$k] = $a;
			}
			catch(SchemaException $ex)
			{
			}
		}
	 }
	 
	/**
	  * Scan class for methods
	  * @param ReflectionClass $refl - object to scan for methods annotations
	  */
	private function scanOperations($refl)
	{
	 	$this->operations = array();
	 	foreach ($refl->getMethods() as $method)
		{
			try
			{
				$k = $method->getName();
				if(substr($k, 0, 1) == '_')
					continue;
					
				$o = new Operation($method);
				$this->operations[$k] = $o;
			}
			catch(SchemaException $ex)
			{
			}
		}
	}

	/**
	 * Scan properties and operations for structures
	 * @param ReflectionClass $refl - object to scan attrubutes and operations for used structures
	 * should be called after scanOperations() and after scanProperties()
	 */
	private function scanStructures($baseClassRefl)
	{
		$this->structures = array();

		foreach($this->properties as $a) 
		{
			if (!$a->isSimple() && !array_key_exists($a->getType(), $this->structures))
				$this->structures[$a->getType()] = NULL;
		}
		
		foreach($this->operations as $o) 
		{
			if ($o->response && $o->response->getType() 
					&& !array_key_exists($o->response->getType(), Schema::$simpleTypes)
					&& !array_key_exists($o->response->getType(), $this->structures))
				$this->structures[$o->response->getType()] = NULL;

			if($o->parameters)
				foreach($o->parameters as $p)
					if (!$p->isSimple() && !array_key_exists($p->getType(), $this->structures))
						$this->structures[$p->getType()] = NULL;
		}
		
		foreach($this->defStructures as $s)
			$this->structures[$s] = NULL;
		
		$this->structures = array_reverse($this->structures);
		
		$missedStructures = array_keys($this->structures);
		do
		{ // Iterrate while we enumirate all structures into depth
			$next = array();
			foreach($missedStructures as $t)
			{
				if ($t == 'self' || $t == 'this' || preg_match('/^(.*)#(.*)$/', $t)) {
					unset($this->structures[$t]);
					continue;
				}
				
				$phpClassName = $t;
				if (strpos($t, '\\') === False && strpos($t, '://') === False) // No namespace defined - use current namespace if any
					if ($baseClassRefl->getNamespaceName())
						$phpClassName = "\\".$baseClassRefl->getNamespaceName()."\\".$t;

				$refl = getAppropriateReflectionClassForObject($phpClassName);
				$this->structures[$t] = new Type($refl);
				foreach($this->structures[$t]->getProperties() as $p)
					if (!$p->isSimple() && !array_key_exists($p->getType(), $this->structures))
						$next[] = $p->getType();
			}
			$missedStructures = $next;
		} while($missedStructures);
		$this->structures = array_reverse($this->structures);
	}
	
	/**
	 * Scan for relations
	 */
	private function scanRelations($refl)
	{
		$this->relations = array();

		foreach ($refl->getProperties() as $prop)
		{
			try
			{
				$k = $prop->getName();
				$a = new Relation($prop);
				$this->relations[$k] = $a;
			}
			catch(SchemaException $ex)
			{
			}
		}
	}
	
	/**
	  * Scan class for meta-data
	  * @param ReflectionClass $refl - object to for basic class annotations
	  */
	private function scanClass($refl)
	{
		$pathToRoot = substr($refl->getFileName(), 0, strpos($refl->getFileName(), 'scripts'));
		if(file_exists($pathToRoot.DIRECTORY_SEPARATOR.'APP-META.xml')) {
			$dom = new \DOMDocument();
			$dom->loadXML(file_get_contents($pathToRoot.DIRECTORY_SEPARATOR.'APP-META.xml'));
			$root = $dom->documentElement;
			$services = $dom->getElementsByTagName('service');
			foreach ($services as $svc) 
			{
				foreach ($svc->childNodes as $child)
				{
					if ($child->nodeType == XML_ELEMENT_NODE && $child->tagName == 'code')
					{
						$relativePath = str_replace('/', DIRECTORY_SEPARATOR, $child->getAttribute('path'));
						$length = strlen($relativePath);
						if(substr($refl->getFileName(), -$length) === $relativePath)
						$this->service = $svc->getAttribute('id');
						break;
					}
				}
			}
			if ($this->service == "") {
				// service is not found in APP-META.xml. Probably runtime modules are under application scripts directory?
				// in this case default it to module name
				$this->service = basename($refl->getFileName(), ".php");
				$this->service = preg_replace("#\.[\d\.]+$#", "", $this->service);
			}
		} else {
			$this->service = basename($refl->getFileName(), ".php");
			$this->service = preg_replace("#\.[\d\.]+$#", "", $this->service);
		}
		
		foreach(getAnnotations($refl) as $ann)
		{
			$name = array_shift($ann);
			if(empty($ann))
				throw new  SchemaException("Property $name has no value");
			switch($name) 
			{
				case "apsVersion":
					$this->apsVersion = $ann[0];
					break;
				case "id":
				case "type": // compat only, will gone 
					$this->id = $ann[0];
					if(preg_match('/\/(\d)+$/', $this->id))
						$this->id .= '.0';
					break;
				case "implements":
					foreach($ann as $impl) {
						$this->implements[] = $impl;
						if(strpos($impl, self::TOP_RESOURCE) !== false) {
							$this->isTopResource = true;
						}
					}
					break;
				case "structure":
					$this->defStructures[] = $ann[0];
					break;
				
				case "access":
					Access::fillAccessByAnnotation($this, $ann[0], $ann[1]);
					break;
				
			}
		}		
	}
	
	/**
	 * 
	 * Scan Class on APS interface definition decorators
	 * Assumed fully qualified name with namespace like \my\favourite\namespace\className
	 * @param string $name - class name to scan
	 */
	public function scan($name)
	{		
		$this->name = $name;
		$refl = getAppropriateReflectionClassForObject($name);
		$this->scanClass($refl);
		$this->scanProperties($refl);
		$this->scanOperations($refl);
		$this->scanStructures($refl);
		$this->scanRelations($refl);
	}
	
	/**
	 * 
	 * Create source of interface for the APS resource by schema
	 */
	public function toSource()
	{		
		$structures = '';
		foreach($this->structures as $structure) {
			$structures .= $structure->toSource();
		}
		$namespace = "namespace ".ltrim(getNamespaceByType($this->id), "\\").";\n";
		$implementsAsStr = '';
		foreach($this->implements as $implements)
			$implementsAsStr .= " * @implements(\"". $implements ."\")\n";
		$defStructuresAsStr = '';
		foreach($this->structures as $structure => $sv)
			$defStructuresAsStr .= " * @structure(". $structure .")\n";
		//----------------------------------------------------------------------
		//Annotation to class generation
		$classShortName = substr($this->name, strrpos($this->name, '\\') + 1);
		$classRepresentation =
		"<?php\n$namespace$structures/**\n * Class $classShortName\n".
		AnnotationUtils::generateQuotedAnnotation($this, 'id') .
		AnnotationUtils::generateSmartAnnotation ($this, 'access') .
		$implementsAsStr .
		$defStructuresAsStr . " */\n";
		//----------------------------------------------------------------------
		//Class name and its parents generation
		$classRepresentation .= "class $classShortName extends \APS\ResourceBase\n{\n";
		//Fields generation
		$sourceProperties = '';
		foreach($this->properties as $property)
		{
			$sourceProperties .= $property->toSource();
		}
		foreach($this->relations as $relation)
		{
			$sourceProperties .= $relation->toSource();
		}
		$classRepresentation .= $sourceProperties;
		$sourceMethods = '';
		foreach($this->operations as $operation) {
			$sourceMethods .= $operation->toSource();
		}
		$classRepresentation .= $sourceMethods; 
		//End of class
		$classRepresentation .= "}\n?>";
		return $classRepresentation;
	}
		
	public function toArray()
	{
		$attrs = array();
		foreach($this->properties as $k => $a) 
			if (!$a->isLink()) 
				$attrs[$k] = $a->toArray();
		
		$ops = array();
		foreach($this->operations as $k => $o) 
				$ops[$k] = $o->toArray();

		$structures = array();
		foreach($this->structures as $k => $t) {
				$structures[$k] = $t->toArray();
		} 
				
		$relations = array();
		foreach($this->relations as $k => $r) {
				$relations[$k] = $r->toArray();
		} 
				
		$ret = array(
    		"apsVersion" => $this->apsVersion,
    		"name" => substr($this->name, strrpos($this->name, '\\') + 1),
    		"id"   => $this->id
		);
		
		if ($this->access)
			$ret["access"] = $this->access->toArray();

		if (!empty($this->implements))
    		$ret["implements"] = $this->implements;
    		
    	if (!empty($attrs))
    		$ret["properties"] = $attrs;
    	
    	if (!empty($ops))
    		$ret["operations"] = $ops;
    		
    	if (!empty($structures))
   			$ret["structures"] = $structures;
   			
   		if (!empty($relations))
			$ret["relations"]  = $relations;

		return $ret;
	}
	
	public function toString($baseOnly = true)
	{
		$schema = ($baseOnly && $this->base) ? $this->base : $this;
		return json_format($schema->toArray());
	}
	
	public function __toString()
	{
		return $this->toString();
	}

	public function fromSource($src, $name) {
		file_put_contents('tmp', $src);
		require_once 'tmp';
		$this->scan($name);
		@unlink('tmp');
	}
	
	/** 
	 * Fill-in schema by JSON-serialized schema definition
	 * @param string $text - JSON representation of schema
	 * @throws SchemaException
	 */
	public function fromString($text)
	{
		$data = json_decode($text);
		if (!$data)
			throw new SchemaException("Can't parse schema structure");

		if ($data->apsVersion != $this->apsVersion)
			throw new SchemaException("Invalid schema version");
			
		
		$this->id = $data->id;
		$this->name = getNamespaceByType($this->id).'\\'.$data->name;
		
		if(isset($data->implements))
			$this->implements = $data->implements;
		
		$this->properties = array();
		if (isset($data->properties))
			foreach($data->properties as $k => $pdata)
				$this->properties[$k] = new Attribute($pdata, getNamespaceByType($this->id), $k);

		$this->operations = array();
		if (isset($data->operations))
			foreach($data->operations as $k => $odata)
				$this->operations[$k] = new Operation($odata, $k);

		$this->structures = array();
		if (isset($data->structures))
			foreach($data->structures as $k => $tdata)				
				$this->structures[$k] = new Type($tdata, getNamespaceByType($this->id), $k);

		$this->relations = array();
		if (isset($data->relations))
			foreach($data->relations as $k => $rdata)				
				$this->relations[$k] = new Relation($rdata, $k);
	}
	
	/**
	 * Merge schema with current schema
	 * @param Schema $schema - schema to be added to current schema
	 * @throws SchemaException
	 */
	public function merge($schema)
	{
		$this->base = clone $this;

		$this->implements = array_unique(array_merge($this->implements, $schema->implements));

		foreach ($schema->properties as $a)
		{
			if (!isset($this->properties[$a->name]))
				$this->properties[$a->name] = clone $a;
		}		

		// Structures are not merged
				
		foreach ($schema->operations as $o)
		{
			// TODO: check for same signature here
			if (!isset($this->operations[$o->name]))
			{
				// $o is not overriden just compy meta
				$this->operations[$o->name] = clone $o;
				continue;
			}
				
			if (isset($this->operations[$o->name]))
				$o1 = $this->operations[$o->name];

			if ($o1)
			{
				if (($o->path != $o1->path) || ($o->verb != $o1->verb))
					throw new SchemaException(
						"Operation '$o->name' defined as '$o->verb $o->path' in $schema->id ".
						"and as $o1->verb $o1->path in $this->id");

				if(!$o1->parameters && $o->parameters)
					$o1->parameters = $o->parameters;
				//else
				//	TODO: check for same parameters
			}
			else
			{
				$this->operations[$o->name] = clone $o;
			}
		}		
	}
}

function getAppropriateReflectionClassForObject($object) {
	if(!class_exists($object)) {
		throw new \Exception("Class $object doesn't exist. Can't load it.");
	}
	if(defined("PHALANGER")) {
		require_once 'aps/2/reflection.php';
		return new ExtendedReflectionClass($object);
	}

	return new \ReflectionClass($object);	
}    

?>
